@@ -1,8 +1,8 @@
 import os
 import sys
 
+import pep440
 import pytest
-from mock import patch
 
 import setupmeta
 import setupmeta.versioning
@@ -12,8 +12,11 @@ from setupmeta.scm import Version
 from . import conftest
 
 
+patch = conftest.mock.patch
+
+
 def new_meta(versioning, name="just-testing", scm=None, setup_py=None, **kwargs):
-    setup_py = setup_py or conftest.resouce("setup.py")
+    setup_py = setup_py or conftest.resource("setup.py")
     upstream = dict(versioning=versioning, scm=scm, _setup_py_path=setup_py)
     if name:
         # Allow to test "missing name" case
@@ -23,6 +26,26 @@ def new_meta(versioning, name="just-testing", scm=None, setup_py=None, **kwargs)
     return SetupMeta().finalize(upstream=upstream)
 
 
+def test_deprecated_strategy_notation():
+    """Custom separators, and the `!` marker will be removed in the future"""
+    with conftest.capture_output() as logged:
+        meta = new_meta("post !", scm=conftest.MockGit(True))
+        versioning = meta.versioning
+        assert str(versioning.strategy) == "branch(main,master):{major}.{minor}.{patch}{post}"
+        check_render(versioning, "1.0.0")
+        assert "PEP-440 allows only '+' as local" in logged
+        assert "'!' character in 'versioning' is now deprecated" in logged
+
+    with conftest.capture_output() as logged:
+        meta = new_meta("distance+!foo", scm=conftest.MockGit(True))
+        versioning = meta.versioning
+        assert str(versioning.strategy) == "branch(main,master):{major}.{minor}.{distance}+foo"
+        check_render(versioning, "1.0.0+foo")
+        check_render(versioning, "1.0.2+foo", distance=2)
+        assert "PEP-440 allows" not in logged
+        assert "'!' character in 'versioning' is now deprecated" in logged
+
+
 def test_disabled():
     with conftest.capture_output():
         meta = new_meta(False)
@@ -85,9 +108,9 @@ def check_render(v, expected, main="1.0", distance=None, cid=None, dirty=False):
 
 
 @patch("setupmeta.model.project_scm", return_value=None)
-def test_no_scm(_):
+def test_no_scm(_, monkeypatch):
     with conftest.capture_output() as logged:
-        fmt = "branch(a,b):{major}.{minor}.{patch}{post} !{.$*FOO*}.{$BAR1*:}{$*BAR2:}{$BAZ:z}{dirty}"
+        fmt = "branch(a,b):{major}.{minor}.{patch}{post}+{.$*FOO*}.{$BAR1*:}{$*BAR2:}{$BAZ:z}{dirty}"
         meta = new_meta(fmt)
         versioning = meta.versioning
 
@@ -103,15 +126,13 @@ def test_no_scm(_):
         assert str(versioning.strategy) == fmt
         assert "BAZ:z" in str(versioning.strategy.extra_bits)
 
-        check_render(versioning, "1.0.0.z")
-        check_render(versioning, "1.0.0.post2.z", distance=2)
-        check_render(versioning, "1.0.0.post2.z.dirty", distance=2, dirty=True)
+        check_render(versioning, "1.0.0+z")
+        check_render(versioning, "1.0.0.post2+z", distance=2)
+        check_render(versioning, "1.0.0.post2+z.dirty", distance=2, dirty=True)
 
-        os.environ["TEST_FOO1"] = "bar"
-        os.environ["TEST_FOO2"] = "baz"
-        check_render(versioning, "1.0.0.post2.bar.z.dirty", distance=2, dirty=True)
-        del os.environ["TEST_FOO1"]
-        del os.environ["TEST_FOO2"]
+        monkeypatch.setenv("TEST_FOO1", "bar")
+        monkeypatch.setenv("TEST_FOO2", "baz")
+        check_render(versioning, "1.0.0.post2+bar.z.dirty", distance=2, dirty=True)
 
         with pytest.raises(setupmeta.UsageError):
             versioning.bump("patch")
@@ -123,16 +144,20 @@ def test_version_from_env_var(*_):
     with conftest.capture_output():
         meta = new_meta("post")
         versioning = meta.versioning
-        assert meta.version == "1.2.3.post4.dirty"
+        assert meta.version == "1.2.3.post4+dirty"
         assert versioning.enabled
         assert not versioning.generate_version_file
         assert not versioning.problem
         assert versioning.scm.is_dirty()
 
 
-def quick_check(versioning, expected, dirty=True, describe="v0.1.2-5-g123"):
+def quick_check(versioning, expected, dirty=True, describe="v0.1.2-5-g123", compliant=True):
     meta = new_meta(versioning, scm=conftest.MockGit(dirty, describe=describe))
     assert meta.version == expected
+    if compliant:
+        main_part, _, _ = meta.version.partition("+")
+        assert pep440.is_canonical(main_part)
+
     versioning = meta.versioning
     assert versioning.enabled
     assert not versioning.generate_version_file
@@ -143,49 +168,74 @@ def quick_check(versioning, expected, dirty=True, describe="v0.1.2-5-g123"):
 @patch.dict(os.environ, {"BUILD_ID": "543"})
 def test_versioning_variants(*_):
     with conftest.capture_output() as logged:
-        quick_check("{major}.{minor}", "0.1.dirty")
+        quick_check("{major}.{minor}", "0.1+dirty")
         quick_check("{major}.{minor}+", "0.1")
-        quick_check("{major}.{minor}-{dirty}", "0.1-.dirty")
-        quick_check("{major}.{minor}{dirty}", "0.1.dirty")
-        quick_check("{major}.{minor}{dirty}+", "0.1.dirty")
+        quick_check("{major}.{minor}{dirty}", "0.1+dirty")
+        quick_check("{major}.{minor}{dirty}+", "0.1+dirty")
         quick_check("{major}.{minor}", "0.1", dirty=False)
+        quick_check("{major}.{minor}+", "0.1", dirty=False)
+        quick_check("{major}.{minor}+", "0.1", dirty=True)
 
-        quick_check("distance", "0.1.5.dirty")
-        quick_check("post", "0.1.2.post5.dirty")
-        quick_check("dev", "0.1.3.dev5.dirty")
+        quick_check("distance", "0.1.5+dirty")
+        quick_check("post", "0.1.2.post5+dirty")
+        quick_check("dev", "0.1.3.dev5+dirty")
+        quick_check("devcommit", "0.1.3.dev5+g123.dirty")
+
+        # Old allowed notations, should remove eventually
+        quick_check("dev+build-id", "0.1.3.dev5+h543.g123.dirty")
+        quick_check("post+build-id", "0.1.2.post5+h543.g123.dirty")
+
+        # Aliases
+        quick_check("changes", "0.1.5+dirty")
+        quick_check("default", "0.1.2.post5+dirty")
+        quick_check("tag", "0.1.2.post5+dirty")
+
+        # Edge cases
+        quick_check("1.2.3", "1.2.3+dirty")
+        quick_check("foo", "foo+dirty", compliant=False)
+
+        quick_check("dev+{commitid}{dirty}", "0.1.3.dev5+g123.dirty")
+        quick_check("dev+{commitid}{dirty}", "0.1.3.dev0+g123.dirty", describe="v0.1.2-0-g123")
+        quick_check("dev+{commitid}{dirty}", "0.1.2+g123", describe="v0.1.2-0-g123", dirty=False)
+
+        quick_check("dev+devcommit", "0.1.3.dev5+g123.dirty")
+        quick_check("dev+devcommit", "0.1.3.dev0+g123.dirty", describe="v0.1.2-0-g123")
+        quick_check("dev+devcommit", "0.1.2", describe="v0.1.2-0-g123", dirty=False)
+
+        quick_check("post+devcommit", "0.1.2.post5+g123.dirty")
+        quick_check("post+devcommit", "0.1.2+g123.dirty", describe="v0.1.2-0-g123")
+        quick_check("post+devcommit", "0.1.2", describe="v0.1.2-0-g123", dirty=False)
 
         quick_check("dev", "0.1.9rc1", dirty=False, describe="v0.1.9-rc.1-0-gebe2789")
         quick_check("devcommit", "0.1.9rc1", dirty=False, describe="v0.1.9-rc.1-0-gebe2789")
-        quick_check("post", "0.1.9rc1.dirty", dirty=True, describe="v0.1.9-rc.1-0-gebe2789")
+        quick_check("post", "0.1.9rc1+dirty", dirty=True, describe="v0.1.9-rc.1-0-gebe2789")
 
         quick_check("dev", "0.1.9rc1.dev1", dirty=False, describe="v0.1.9-rc.1-1-gebe2789")
-        quick_check("devcommit", "0.1.9rc1.dev1-gebe2789", dirty=False, describe="v0.1.9-rc.1-1-gebe2789")
-        quick_check("devcommit", "0.1.9rc1.dev1-gebe2789-dirty", dirty=True, describe="v0.1.9-rc.1-1-gebe2789")
+        quick_check("devcommit", "0.1.9rc1.dev1+gebe2789", dirty=False, describe="v0.1.9-rc.1-1-gebe2789")
+        quick_check("devcommit", "0.1.9rc1.dev1+gebe2789.dirty", dirty=True, describe="v0.1.9-rc.1-1-gebe2789")
         quick_check("post", "0.1.9rc1.post1", dirty=False, describe="v0.1.9-rc.1-1-gebe2789")
 
-        quick_check("devcommit", "0.1.3.dev5-g123", dirty=False)
-        quick_check("devcommit", "0.1.3.dev5-g123-dirty")
-        quick_check("tag dev", "0.1.3.dev5.dirty")
-        quick_check("tag+dev", "0.1.3.dev5+.dirty")
+        quick_check("devcommit", "0.1.3.dev5+g123", dirty=False)
+        quick_check("devcommit", "0.1.3.dev5+g123.dirty")
         quick_check("build-id", "0.1.5+h543.g123.dirty")
-        quick_check("dev+build-id", "0.1.3.dev5+h543.g123.dirty")
 
         quick_check("post", "5.0.0a1.post1", dirty=False, describe="v5.0-a.1-1-gebe2789")
-        quick_check("post", "5.0.0a1.rc2.post7", dirty=False, describe="v5.a1rc2-7-gebe2789")
+        quick_check("post", "5.0.0a1.rc2.post7", dirty=False, describe="v5.a1rc2-7-gebe2789", compliant=False)
         quick_check("dev", "0.1.0a0.dev8", dirty=False, describe="v0.1.a-8-gebe2789")
 
         # Patch is not bumpable
-        quick_check("dev", "0.1.0rc0.dev5.dirty", describe="v0.1.rc-5-g123")
-        quick_check("dev", "0.1.0rc1.dev5.dirty", describe="v0.1.rc1-5-g123")
-        quick_check("dev", "0.1.0rc1.dev5.dirty", describe="v0.1.rc.1-5-g123")
-        quick_check("dev", "0.1.0rc1.dev5.dirty", describe="v0.1.rc-1-5-g123")
+        quick_check("dev", "0.1.0rc0.dev5+dirty", describe="v0.1.rc-5-g123")
+        quick_check("dev", "0.1.0rc1.dev5+dirty", describe="v0.1.rc1-5-g123")
+        quick_check("dev", "0.1.0rc1.dev5+dirty", describe="v0.1.rc.1-5-g123")
+        quick_check("dev", "0.1.0rc1.dev5+dirty", describe="v0.1.rc-1-5-g123")
 
         # On tag
         quick_check("dev", "0.1.2", describe="v0.1.2-0-g123", dirty=False)
-        quick_check("dev", "0.1.3.dev0.dirty", describe="v0.1.2-0-g123", dirty=True)
+        quick_check("dev", "0.1.3.dev0+dirty", describe="v0.1.2-0-g123", dirty=True)
         quick_check("devcommit", "0.1.2", describe="v0.1.2", dirty=False)
         quick_check("devcommit", "0.1.2", describe="v0.1.2-0-g123", dirty=False)
-        quick_check("devcommit", "0.1.3.dev0-g123-dirty", describe="v0.1.2-0-g123", dirty=True)
+        quick_check("devcommit", "0.1.3.dev7+g123", describe="v0.1.2-7-g123", dirty=False)
+        quick_check("devcommit", "0.1.3.dev0+g123.dirty", describe="v0.1.2-0-g123", dirty=True)
 
         assert "patch version component should be .0" in logged
 
@@ -202,13 +252,20 @@ def test_bump_patch():
 
 def test_no_extra():
     with conftest.capture_output() as logged:
+        meta = new_meta("{major}.{minor}+", scm=conftest.MockGit(True))
+        versioning = meta.versioning
+        assert str(versioning.strategy) == "branch(main,master):{major}.{minor}"
+        check_render(versioning, "1.0")
+        check_render(versioning, "1.0", distance=2)
+        check_render(versioning, "1.0", distance=2, dirty=True)
+
         meta = new_meta("{major}.{minor}.{$FOO}+", scm=conftest.MockGit(True))
         versioning = meta.versioning
-        assert meta.version == "0.1.None"
-        assert str(versioning.strategy) == "branch(main,master):{major}.{minor}.{$FOO}+"
-        check_render(versioning, "1.0.None")
-        check_render(versioning, "1.0.None", distance=2)
-        check_render(versioning, "1.0.None", distance=2, dirty=True)
+        assert meta.version == "0.1+None"
+        assert str(versioning.strategy) == "branch(main,master):{major}.{minor}+{$FOO}"
+        check_render(versioning, "1.0+None")
+        check_render(versioning, "1.0+None", distance=2)
+        check_render(versioning, "1.0+None", distance=2, dirty=True)
 
         assert "patch version component should be .0" in logged
 
@@ -216,23 +273,25 @@ def test_no_extra():
 def extra_version(version):
     if version.dirty:
         return "extra"
+
     if version.distance:
         return "d%s" % version.distance
+
     return ""
 
 
 def test_invalid_part():
     with conftest.capture_output() as logged:
-        versioning = dict(foo="bar", main="{foo}.{major}.{minor}{", extra=extra_version, separator="-")
+        versioning = dict(foo="bar", main="{foo}.{major}.{minor}{", extra=extra_version)
         meta = new_meta(versioning, scm=conftest.MockGit())
         versioning = meta.versioning
         assert "invalid" in str(versioning.strategy.main_bits)
         assert meta.version is None
         assert versioning.problem == "invalid versioning part 'foo'"
-        assert str(versioning.strategy) == "branch(main,master):{foo}.{major}.{minor}{-function 'extra_version'"
+        assert str(versioning.strategy) == "branch(main,master):{foo}.{major}.{minor}{+function 'extra_version'"
         check_render(versioning, "invalid.1.0")
-        check_render(versioning, "invalid.1.0-d2", distance=2)
-        check_render(versioning, "invalid.1.0-extra", distance=2, dirty=True)
+        check_render(versioning, "invalid.1.0+d2", distance=2)
+        check_render(versioning, "invalid.1.0+extra", distance=2, dirty=True)
 
         assert "Ignored fields for 'versioning': {'foo': 'bar'}" in logged
 
@@ -245,9 +304,9 @@ def test_invalid_part():
 
 def test_invalid_main():
     with conftest.capture_output() as logged:
-        meta = new_meta(dict(main=extra_version, extra="", separator=" "), scm=conftest.MockGit())
+        meta = new_meta(dict(main=extra_version, extra=""), scm=conftest.MockGit())
         versioning = meta.versioning
-        assert str(versioning.strategy) == "branch(main,master):function 'extra_version' "
+        assert str(versioning.strategy) == "branch(main,master):function 'extra_version'"
         check_render(versioning, "")
         check_render(versioning, "d2", distance=2)
         check_render(versioning, "extra", distance=2, dirty=True)
@@ -274,29 +333,17 @@ def test_distance_marker():
         assert versioning.enabled
         assert not versioning.problem
         assert not versioning.strategy.problem
-        assert meta.version == "0.1.3.dirty"
-        assert str(versioning.strategy) == "branch(main,master):{major}.{minor}.{distance}{dirty}"
+        assert meta.version == "0.1.3+dirty"
+        assert str(versioning.strategy) == "branch(main,master):{major}.{minor}.{distance}+{dirty}"
 
 
-@patch.dict(os.environ, {"BUILD_ID": "543"})
-def test_preconfigured_build_id(*_):
+def test_preconfigured_build_id():
     """Verify that short notations expand to the expected format"""
-    check_preconfigured("branch(main,master):{major}.{minor}.{patch}{post}{dirty}", "post", "default")
-
-    check_preconfigured("branch(main,master):{major}.{minor}.{distance}{dirty}", "distance")
-
-    check_preconfigured(
-        "branch(main,master):{major}.{minor}.{distance}+!h{$*BUILD_ID:local}.{commitid}{dirty}",
-        "build-id",
-        "distance+build-id"
-    )
-
-    check_preconfigured(
-        "branch(main,master):{major}.{minor}.{patch}{post}+!h{$*BUILD_ID:local}.{commitid}{dirty}",
-        "+build-id",
-        "default+build-id",
-        "post+build-id",
-    )
+    check_preconfigured("branch(main,master):{major}.{minor}.{patch}{post}+{dirty}", "post", "default", "tag")
+    check_preconfigured("branch(main,master):{major}.{minor}.{distance}+{dirty}", "distance", "changes")
+    check_preconfigured("branch(main,master):{major}.{minor}.{distance}+h{$*BUILD_ID:local}.{commitid}{dirty}", "build-id")
+    check_preconfigured("branch(main,master):{major}.{minor}.{patch}{dev}+h{$*BUILD_ID:local}.{commitid}{dirty}", "dev+build-id")
+    check_preconfigured("branch(main,master):{major}.{minor}.{patch}{post}+h{$*BUILD_ID:local}.{commitid}{dirty}", "post+build-id")
 
 
 def check_preconfigured(expected, *shorts):
@@ -327,9 +374,9 @@ def check_strategy_distance(dirty):
     assert not versioning.strategy.problem
     assert "major" in str(versioning.strategy.main_bits)
     assert "dirty" in str(versioning.strategy.extra_bits)
-    assert str(versioning.strategy) == "branch(main,master):{major}.{minor}.{distance}{dirty}"
+    assert str(versioning.strategy) == "branch(main,master):{major}.{minor}.{distance}+{dirty}"
     if dirty:
-        assert meta.version == "0.1.3.dirty"
+        assert meta.version == "0.1.3+dirty"
 
         with pytest.raises(setupmeta.UsageError):
             # Can't effectively bump if checkout is dirty
@@ -353,7 +400,7 @@ def check_strategy_build_id(dirty):
     assert not versioning.strategy.problem
     assert "major" in str(versioning.strategy.main_bits)
     assert "commitid" in str(versioning.strategy.extra_bits)
-    assert str(versioning.strategy) == "branch(main,master):{major}.{minor}.{distance}+!h{$*BUILD_ID:local}.{commitid}{dirty}"
+    assert str(versioning.strategy) == "branch(main,master):{major}.{minor}.{distance}+h{$*BUILD_ID:local}.{commitid}{dirty}"
     if dirty:
         assert meta.version == "0.1.3+h543.g123.dirty"
 
@@ -425,7 +472,7 @@ def test_brand_new_project():
 
         # Now stage a file
         conftest.run_git("add", "setup.py")
-        check_version_output("0.0.0.dirty")
+        check_version_output("0.0.0+dirty")
 
         # Unstage it
         conftest.run_git("reset", "setup.py")
@@ -465,12 +512,12 @@ def test_git_versioning(sample_project):
     # Modify existing file makes checkout dirty
     write_to_file("sample.py", "print('hello')")
     output = setupmeta.run_program(sys.executable, "setup.py", "--version", capture=True)
-    assert output == "0.1.0.dirty"
+    assert output == "0.1.0+dirty"
 
     # git add -> version should still be dirty, as we didn't commit yet
     conftest.run_git("add", "sample.py")
     output = setupmeta.run_program(sys.executable, "setup.py", "--version", capture=True)
-    assert output == "0.1.0.dirty"
+    assert output == "0.1.0+dirty"
 
     # git commit -> version reflects new distance
     conftest.run_git("commit", "-m", "Testing")
